#!/usr/bin/env python

import os
import glob
import pandas as pd
from collections import OrderedDict
from reframed import Environment
from .smetana import mip_score, mro_score, sc_score, mp_score, mu_score, minimal_environment
from random import sample
from reframed.io.cache import ModelCache
from smetana.legacy import Community
from math import inf


def extract_id_from_filepath(filepath):
    filename = os.path.basename(filepath)

    if filename.endswith('.xml'):
        organism_id = filename[:-4]
    elif filename.endswith('.xml.gz'):
        organism_id = filename[:-7]
    else:
        raise IOError('Unrecognized extension in file {}. Valid extensions are .xml and .xml.gz'.format(filename))

    return organism_id


def build_cache(models, flavor=None):
    ids = [extract_id_from_filepath(model) for model in models]

    if not flavor:
        flavor = 'fbc2'

    load_args = {'flavor': flavor}

    def post_process(model):
        if 'R_ATPM' in model.reactions:
            model.reactions.R_ATPM.lb = 0

    return ModelCache(ids, models, load_args=load_args, post_processing=post_process)


def load_communities(models, communities, other, flavor):
    if len(models) == 1 and '*' in models[0]:
        pattern = models[0]
        models = glob.glob(pattern)
        if len(models) == 0:
            raise RuntimeError("No files found: {}".format(pattern))

    if other is not None:
        df = pd.read_csv(other, header=None)
        other_models = set(df[0])
    else:
        other_models = set()

    model_cache = build_cache(models, flavor)

    if communities is not None:
        df = pd.read_csv(communities, sep='\t', header=None, dtype=str)
        comm_dict = OrderedDict((name, group[1].tolist()) for name, group in df.groupby(0))
    else:
        comm_dict = {'all': model_cache.get_ids()}

    if other:
        missing = other_models - set(model_cache.get_ids())
        if missing:
            print("Warning: {} models in perturbation file missing.".format(len(missing)))

    return model_cache, comm_dict, other_models


def load_media_db(filename, sep='\t', medium_col='medium', compound_col='compound'):
    """ Load media library file. """

    data = pd.read_csv(filename, sep=sep)
    media_db = data[[medium_col, compound_col]].groupby(medium_col).agg(lambda x: list(x))

    return media_db[compound_col].to_dict()


def load_media(media, mediadb, exclude, other):
    media_db = None

    if media:
        if mediadb is None:
            raise IOError('Please provide media library with --mediadb option.')
        else:
            media_db = load_media_db(mediadb)
        media = media.split(',')
    else:
        media = [None]

    if exclude:
        df = pd.read_csv(exclude, header=None)
        excluded_mets = set(df[0])
    else:
        excluded_mets = set()

    if other:
        df = pd.read_csv(other, header=None)
        other_mets = set(df[0])
    else:
        other_mets = set()

    return media, media_db, excluded_mets, other_mets


def define_environment(medium, media_db, community, mode, aerobic, verbose, min_mol_weight, use_lp):
    max_uptake = 10.0 * len(community.organisms)

    if medium:
        fmt_func = lambda x: "R_EX_M_{}_e_pool".format(x)
        env = Environment.from_compounds(media_db[medium], fmt_func=fmt_func, max_uptake=max_uptake)
        medium_id = medium
    elif mode == "global":
        env = Environment.complete(community.merged, max_uptake=max_uptake)
        medium_id = 'complete'

        if aerobic is not None and aerobic:
            env["R_EX_M_o2_e_pool"] = (-max_uptake, inf)

        if aerobic is not None and not aerobic:
            env["R_EX_M_o2_e_pool"] = (0, inf)

    else:
        env = minimal_environment(community, aerobic, verbose=verbose, min_mol_weight=min_mol_weight,
                                  use_lp=use_lp, max_uptake=max_uptake)
        medium_id = "minimal"

    return medium_id, env


def run_global(comm_id, community, organisms, medium_id, excluded_mets, env, verbose, min_mol_weight, use_lp, debug):
    global_data = []
    debug_data = []

    if verbose:
        print('Running MIP for community {} on medium {}...'.format(comm_id, medium_id))

    mip, extras = mip_score(community, environment=env, verbose=verbose,
                            min_mol_weight=min_mol_weight, use_lp=use_lp, exclude=excluded_mets)

    if mip is None:
        mip = 'n/a'

    if debug and extras is not None:
        mip_ni = ','.join(sorted(extras['noninteracting_medium']))
        mip_i = ','.join(sorted(extras['interacting_medium']))
        debug_data.append((comm_id, medium_id, 'mip', 'ni', mip_ni))
        debug_data.append((comm_id, medium_id, 'mip', 'i', mip_i))

    if verbose:
        print('Running MRO for community {} on medium {}...'.format(comm_id, medium_id))

    mro, extras = mro_score(community, environment=env, verbose=verbose,
                            min_mol_weight=min_mol_weight, use_lp=use_lp, exclude=excluded_mets)

    if mro is None:
        mro = 'n/a'

    if debug and extras is not None:
        comm_medium = ','.join(sorted(extras['community_medium']))
        debug_data.append((comm_id, medium_id, 'mro', 'community', comm_medium))
        for org, values in extras['individual_media'].items():
            org_medium = ','.join(sorted(values))
            debug_data.append((comm_id, medium_id, 'mro', org, org_medium))

    global_data.append((comm_id, medium_id, len(organisms), mip, mro))

    return global_data, debug_data


def run_detailed(comm_id, community, medium_id, excluded_mets, env, verbose, min_mol_weight, ignore_coupling):
    smt_data = []

    exclude_bigg = {'M_{}_e'.format(x) for x in excluded_mets}

    if not ignore_coupling:
        if verbose:
            print('Running SCS for community {} on medium {}...'.format(comm_id, medium_id))

        scs = sc_score(community, environment=env, verbose=verbose)

    if verbose:
        print('Running MUS for community {} on medium {}...'.format(comm_id, medium_id))

    mus = mu_score(community, environment=env, verbose=verbose,
                   min_mol_weight=min_mol_weight)

    if verbose:
        print('Running MPS for community {} on medium {}...'.format(comm_id, medium_id))

    mps = mp_score(community, environment=env)

    pairs = [(org1, org2) for org1 in community.organisms
             for org2 in community.organisms if org1 != org2]

    for org1, org2 in pairs:
        if not ignore_coupling and scs[org1] is None:
            continue
        if mus[org1] is None:
            continue
        if mps[org2] is None:
            continue

        metabolites = (set(mus[org1]) | set(mps[org2])) - exclude_bigg

        for met in sorted(metabolites):
            mus_o1_met = mus[org1].get(met, 0)
            mps_o2_met = mps[org2].get(met, 0)

            if ignore_coupling:
                scs_o1_o2 = 'n/a'
                smt = mus_o1_met * mps_o2_met
            else:
                scs_o1_o2 = scs[org1][org2]
                smt = scs_o1_o2 * mus_o1_met * mps_o2_met
            smt_data.append((comm_id, medium_id, org1, org2, met, scs_o1_o2, mus_o1_met, mps_o2_met, smt))

    return smt_data


def run_abiotic(comm_id, sense, community, medium_id, excluded_mets, env, verbose, min_mol_weight, other_mets, n, p,
                ignore_coupling):

    medium = set(env.get_compounds(fmt_func=lambda x: x[7:-7]))
    max_uptake = 10.0 * len(community.organisms)

    if sense == 'add':
        modified = sorted(other_mets - (medium | excluded_mets))

    if sense == 'rm':
        modified = sorted(other_mets - (medium | excluded_mets))
        n_extra_cpds = 2*p
        modified = sample(modified, n_extra_cpds)
        medium = medium | set(modified)
        env = Environment.from_compounds(medium, fmt_func=lambda x: f"R_EX_M_{x}_e_pool",  max_uptake=max_uptake)

    if len(modified) < p:
        raise RuntimeError("Insufficient compounds ({}) to perform ({}) perturbations.".format(len(modified), p))

    if n == 0:
        do_all = True
        n = len(modified)
        if verbose:
            print('Running {} systematic abiotic perturbations with 1 compound...'.format(n))
    else:
        do_all = False
        if verbose:
            print('Running {} random abiotic perturbations with {} compounds...'.format(n, p))

    data = run_detailed(comm_id, community, medium_id, excluded_mets, env, False, min_mol_weight, ignore_coupling)

    for i in range(n):
        if do_all:
            if sense == 'add':
                new_compounds = list(medium) + [modified[i]]
            if sense == 'rm':
                new_compounds = medium - {modified[i]}
            new_id = "{}_{}".format(medium_id, modified[i])
        else:
            if sense == 'add':
                new_compounds = list(medium) + sample(modified, p)
            if sense == 'rm':
                new_compounds = medium - set(sample(modified, p))
            new_id = "{}_{}".format(medium_id, i + 1)

        new_env = Environment.from_compounds(new_compounds, fmt_func=lambda x: f"R_EX_M_{x}_e_pool",
                                             max_uptake=max_uptake)
        entries = run_detailed(comm_id, community, new_id, excluded_mets, new_env, False, min_mol_weight,
                               ignore_coupling)
        data.extend(entries)

    return data


def run_biotic(comm_id, community, medium_id, excluded_mets, env, verbose, min_mol_weight, other_models, model_cache,
               n, p, ignore_coupling):
    inserted = sorted(other_models - set(community.organisms))

    if len(inserted) < p:
        raise RuntimeError("Insufficient species ({}) to perform ({}) perturbations.".format(len(inserted), p))

    if n == 0:
        do_all = True
        n = len(inserted)
        if verbose:
            print('Running {} systematic biotic perturbations with 1 species...'.format(n))

    else:
        do_all = False
        if verbose:
            print('Running {} random biotic perturbations with {} species...'.format(n, p))

    data = run_detailed(comm_id, community, medium_id, excluded_mets, env, False, min_mol_weight, ignore_coupling)

    for i in range(n):
        if do_all:
            new_species = list(community.organisms) + [inserted[i]]
            new_id = "{}_{}".format(comm_id, inserted[i])
        else:
            new_species = list(community.organisms) + sample(inserted, p)
            new_id = "{}_{}".format(comm_id, i + 1)

        comm_models = [model_cache.get_model(org_id, reset_id=True) for org_id in new_species]
        new_community = Community(comm_id, comm_models, copy_models=False, create_biomass=False)
        entries = run_detailed(new_id, new_community, medium_id, excluded_mets, env, False, min_mol_weight,
                               ignore_coupling)
        data.extend(entries)

    return data


def export_results(mode, output, data, debug_data, zeros):
    prefix = output + '_' if output else ''

    if mode == "global":

        df = pd.DataFrame(data, columns=['community', 'medium', 'size', "mip", "mro"])
        df.to_csv(prefix + 'global.tsv', sep='\t', index=False)

        if len(debug_data) > 0:
            df = pd.DataFrame(debug_data, columns=['community', 'medium', 'key1', "key2", "data"])
            df.to_csv(prefix + 'debug.tsv', sep='\t', index=False)

    else:

        columns = ['community', 'medium', 'receiver', 'donor', 'compound', 'scs', 'mus', 'mps', 'smetana']
        df = pd.DataFrame(data, columns=columns)

        if not zeros:
            df = df.query('smetana > 0')

        df.to_csv(prefix + 'detailed.tsv', sep='\t', index=False)


def main(models, communities=None, mode=None, output=None, flavor=None, media=None, mediadb=None, aerobic=None,
         zeros=False,verbose=False, min_mol_weight=False, use_lp=False, exclude=None, debug=False,
         other=None, n=1, p=1, ignore_coupling=False):

    other_models = other if mode == "biotic" else None
    model_cache, comm_dict, other_models = load_communities(models, communities, other_models, flavor)

    other_mets = other if mode == "abiotic" or mode == 'abiotic-rm' else None
    media, media_db, excluded_mets, other_mets = load_media(media, mediadb, exclude, other_mets)

    data = []
    debug_data = []

    for comm_id, organisms in list(comm_dict.items()):

        if verbose:
            print("Loading community: " + comm_id)

        comm_models = [model_cache.get_model(org_id, reset_id=True) for org_id in organisms]
        community = Community(comm_id, comm_models, copy_models=False)

        for medium in media:

            medium_id, env = define_environment(medium, media_db, community, mode, aerobic, verbose, min_mol_weight, use_lp)

            if mode == "global":
                entries, debug_entries = run_global(comm_id, community, organisms, medium_id, excluded_mets, env,
                                                    verbose, min_mol_weight, use_lp, debug)

            if mode == "detailed":
                entries = run_detailed(comm_id, community, medium_id, excluded_mets, env, verbose, min_mol_weight,
                                       ignore_coupling)

            if mode == "abiotic":
                entries = run_abiotic(comm_id, 'add', community, medium_id, excluded_mets, env, verbose, min_mol_weight,
                                      other_mets, n, p, ignore_coupling)

            if mode == "abiotic-rm":
                entries = run_abiotic(comm_id, 'rm', community, medium_id, excluded_mets, env, verbose, min_mol_weight,
                                       other_mets, n, p, ignore_coupling)

            if mode == "biotic":
                entries = run_biotic(comm_id, community, medium_id, excluded_mets, env, verbose, min_mol_weight,
                                     other_models, model_cache, n, p, ignore_coupling)

            data.extend(entries)

            if debug:
                debug_data.extend(debug_entries)

    export_results(mode, output, data, debug_data, zeros)

    if verbose:
        print('Done.')




from reframed import minimal_medium, solver_instance, Environment
from reframed.solvers.solver import VarType
from reframed.solvers.solution import Status

from collections import Counter
from itertools import combinations, chain
from warnings import warn
from math import isinf, inf


def sc_score(community, environment=None, min_growth=0.1, n_solutions=100, verbose=True, abstol=1e-6, use_pool=False):
    """
    Calculate frequency of community species dependency on each other

    Zelezniak A. et al, Metabolic dependencies drive species co-occurrence in diverse microbial communities (PNAS 2015)

    Args:
        community (Community): microbial community
        environment (Environment): metabolic environment (optional)
        min_growth (float): minimum growth rate (default: 0.1)
        abstol (float): tolerance for detecting a non-zero exchange flux (default: 1e-6)
        n_solutions (int): number of alternative solutions to calculate (default: 100)

    Returns:
        dict: Keys are dependent organisms, values are dictionaries with required organism frequencies
    """

    community = community.copy(copy_models=False, interacting=True, create_biomass=False,
                               merge_extracellular_compartments=False)

    if environment:
        environment.apply(community.merged, inplace=True, warning=False)

    for b in community.organisms_biomass_reactions.values():
        community.merged.reactions[b].lb = 0

    solver = solver_instance(community.merged)

    for org_id in community.organisms:
        org_var = 'y_{}'.format(org_id)
        solver.add_variable(org_var, 0, 1, vartype=VarType.BINARY)

    solver.update()

    bigM = 1000
    for org_id, rxns in community.organisms_reactions.items():
        org_var = 'y_{}'.format(org_id)
        for r_id in rxns:
            if r_id == community.organisms_biomass_reactions[org_id]:
                continue
            solver.add_constraint('c_{}_lb'.format(r_id), {r_id: 1, org_var: bigM}, '>', 0)
            solver.add_constraint('c_{}_ub'.format(r_id), {r_id: 1, org_var: -bigM}, '<', 0)

    solver.update()

    scores = {}

    for org_id, biomass_id in community.organisms_biomass_reactions.items():
        other = {o for o in community.organisms if o != org_id}
        solver.add_constraint('SMETANA_Biomass', {community.organisms_biomass_reactions[org_id]: 1}, '>', min_growth)
        objective = {"y_{}".format(o): 1.0 for o in other}

        if not use_pool:
            previous_constraints = []
            donors_list = []
            failed = False

            for i in range(n_solutions):
                sol = solver.solve(objective, minimize=True, get_values=list(objective.keys()))

                if sol.status != Status.OPTIMAL:
                    failed = i == 0
                    break

                donors = [o for o in other if sol.values["y_{}".format(o)] > abstol]
                donors_list.append(donors)

                previous_con = 'iteration_{}'.format(i)
                previous_constraints.append(previous_con)
                previous_sol = {"y_{}".format(o): 1 for o in donors}
                solver.add_constraint(previous_con, previous_sol, '<', len(previous_sol) - 1)

            for constr_id in ['SMETANA_Biomass'] + previous_constraints:
                solver.remove_constraint(constr_id)

            if not failed:
                donors_list_n = float(len(donors_list))
                donors_counter = Counter(chain(*donors_list))
                scores[org_id] = {o: donors_counter[o] / donors_list_n for o in other}
            else:
                if verbose:
                    warn('SCS: Failed to find a solution for growth of ' + org_id)
                scores[org_id] = None

        else:
            sols = solver.solve(objective, minimize=True, get_values=list(objective.keys()),
                                pool_size=n_solutions, pool_gap=0.5)
            solver.remove_constraint('SMETANA_Biomass')

            if len(sols) == 0:
                scores[org_id] = None
                if verbose:
                    warn('SCS: Failed to find a solution for growth of ' + org_id)
            else:
                donor_count = [o for sol in sols for o in other if sol.values["y_{}".format(o)] > abstol]
                donor_count = Counter(donor_count)
                scores[org_id] = {o: donor_count[o] / len(sols) for o in other}

    return scores


def mu_score(community, environment=None, min_mol_weight=False, min_growth=0.1, max_uptake=10.0,
             abstol=1e-6, validate=False, n_solutions=100, pool_gap=0.5, verbose=True):
    """
    Calculate frequency of metabolite requirement for species growth

    Zelezniak A. et al, Metabolic dependencies drive species co-occurrence in diverse microbial communities (PNAS 2015)

    Args:
        community (Community): microbial community
        environment (Environment): metabolic environment
        min_mol_weight (bool): Prefer smaller compounds (default: False)
        min_growth (float): minimum growth rate (default: 0.1)
        max_uptake (float): maximum uptake rate (default: 10)
        abstol (float): tolerance for detecting a non-zero exchange flux (default: 1e-6)
        validate (bool): validate solution using FBA (for debugging purposes, default: False)
        n_solutions (int): number of alternative solutions to calculate (default: 100)

    Returns:
        dict: Keys are organism names, values are dictionaries with metabolite frequencies 
        dict: Extra information
    """

    if environment:
        environment.apply(community.merged, inplace=True, warning=False)

    max_uptake = max_uptake * len(community.organisms)
    scores = {}
    solver = solver_instance(community.merged)

    for org_id in community.organisms:
        exchange_rxns = community.organisms_exchange_reactions[org_id]
        biomass_reaction = community.organisms_biomass_reactions[org_id]
        community.merged.biomass_reaction = biomass_reaction

        medium_list, sols = minimal_medium(community.merged, exchange_reactions=list(exchange_rxns.keys()),
                                           min_mass_weight=min_mol_weight, min_growth=min_growth,
                                           n_solutions=n_solutions, max_uptake=max_uptake, validate=validate,
                                           abstol=abstol, use_pool=False, pool_gap=pool_gap, solver=solver,
                                           warnings=False)

        if medium_list:
            counter = Counter(chain(*medium_list))

            scores[org_id] = {cnm.original_metabolite: counter[ex] / len(medium_list)
                              for ex, cnm in exchange_rxns.items()}
        else:
            if verbose:
                warn('MUS: Failed to find a minimal growth medium for ' + org_id)
            scores[org_id] = None

    return scores


def mp_score(community, environment=None, abstol=1e-3):
    """
    Discover metabolites which species can produce in community

    Zelezniak A. et al, Metabolic dependencies drive species co-occurrence in diverse microbial communities (PNAS 2015)

    Args:
        community (Community): community object
        environment (Environment): Metabolic environment in which the SMETANA score is colulated
        min_growth (float): minimum growth rate (default: 0.1)
        max_uptake (float): maximum uptake rate (default: 10)
        abstol (float): tolerance for detecting a non-zero exchange flux (default: 1e-6)

    Returns:
        dict: Keys are model names, values are list with produced compounds
        dict: Extra information
    """

    if environment:
        environment.apply(community.merged, inplace=True, warning=False)
        env_compounds = environment.get_compounds(fmt_func=lambda x: x[5:-5])
    else:
        env_compounds = set()

    for exchange_rxns in community.organisms_exchange_reactions.values():
        for r_id in exchange_rxns.keys():
            rxn = community.merged.reactions[r_id]
            if isinf(rxn.ub):
                rxn.ub = 1000

    solver = solver_instance(community.merged)

    scores = {}

    for org_id, exchange_rxns in community.organisms_exchange_reactions.items():
        scores[org_id] = {}

        remaining = [r_id for r_id, cnm in exchange_rxns.items() if cnm.original_metabolite not in env_compounds]

        while len(remaining) > 0:
            sol = solver.solve(objective={r_id: 1 for r_id in remaining}, minimize=False, get_values=remaining)

            if sol.status != Status.OPTIMAL:
                break

            blocked = [r_id for r_id in remaining if sol.values[r_id] < abstol]

            if len(blocked) == len(remaining):
                break

            for r_id in remaining:
                if sol.values[r_id] >= abstol:
                    cnm = exchange_rxns[r_id]
                    scores[org_id][cnm.original_metabolite] = 1

            remaining = blocked

        for r_id in remaining:
            sol = solver.solve(objective={r_id: 1}, minimize=False, get_values=False)
            cnm = exchange_rxns[r_id]

            if sol.status == Status.OPTIMAL and sol.fobj > abstol:
                scores[org_id][cnm.original_metabolite] = 1
            else:
                scores[org_id][cnm.original_metabolite] = 0

    return scores


def mip_score(community, environment=None, min_mol_weight=False, min_growth=0.1, direction=-1, max_uptake=10,
              validate=False, verbose=True, use_lp=False, exclude=None):
    """
    Implements the metabolic interaction potential (MIP) score as defined in (Zelezniak et al, 2015).

    Args:
        community (Community): microbial community model
        environment (Environment): Metabolic environment in which the SMETANA score is calculated
        direction (int): direction of uptake reactions (negative or positive, default: -1)
        extracellular_id (str): extracellular compartment id
        min_mol_weight (bool): minimize by molecular weight of nutrients (default: False)
        min_growth (float): minimum growth rate (default: 0.1)
        max_uptake (float): maximum uptake rate (default: 10)
        validate (bool): validate solution using FBA (for debugging purposes, default: False)

    Returns:
        float: MIP score
    """

    noninteracting = community.copy(copy_models=False, interacting=False)
    exch_reactions = set(community.merged.get_exchange_reactions())
    max_uptake = max_uptake * len(community.organisms)

    if environment:
        environment.apply(noninteracting.merged, inplace=True, warning=False)
        exch_reactions &= set(environment)

    noninteracting_medium, sol1 = minimal_medium(noninteracting.merged, exchange_reactions=exch_reactions,
                                                 direction=direction, min_mass_weight=min_mol_weight,
                                                 min_growth=min_growth, max_uptake=max_uptake, validate=validate,
                                                 warnings=False, milp=(not use_lp))
    if noninteracting_medium is None:
        if verbose:
            warn('MIP: Failed to find a valid solution for non-interacting community')
        return None, None

    # anabiotic environment is limited to non-interacting community minimal media
    noninteracting_env = Environment.from_reactions(noninteracting_medium, max_uptake=max_uptake)
    noninteracting_env.apply(community.merged, inplace=True)

    interacting_medium, sol2 = minimal_medium(community.merged, direction=direction, exchange_reactions=noninteracting_medium,
                                              min_mass_weight=min_mol_weight, min_growth=min_growth, milp=(not use_lp),
                                              max_uptake=max_uptake, validate=validate, warnings=False)

    if interacting_medium is None:
        if verbose:
            warn('MIP: Failed to find a valid solution for interacting community')
        return None, None

    if exclude is not None:
        exclude_rxns = {'R_EX_M_{}_e_pool'.format(x) for x in exclude}
        interacting_medium = set(interacting_medium) - exclude_rxns
        noninteracting_medium = set(noninteracting_medium) - exclude_rxns

    score = len(noninteracting_medium) - len(interacting_medium)

    noninteracting_medium = [r_id[7:-7] for r_id in noninteracting_medium]
    interacting_medium = [r_id[7:-7] for r_id in interacting_medium]

    extras = {
        'noninteracting_medium': noninteracting_medium,
        'interacting_medium': interacting_medium
    }

    return score, extras


def mro_score(community, environment=None, direction=-1, min_mol_weight=False, min_growth=0.1, max_uptake=10,
              validate=False, verbose=True, use_lp=False, exclude=None):
    """
    Implements the metabolic resource overlap (MRO) score as defined in (Zelezniak et al, 2015).

    Args:
        community (Community): microbial community model
        environment (Environment): Metabolic environment in which the SMETANA score is colulated
        direction (int): direction of uptake reactions (negative or positive, default: -1)
        extracellular_id (str): extracellular compartment id
        min_mol_weight (bool): minimize by molecular weight of nutrients (default: False)
        min_growth (float): minimum growth rate (default: 0.1)
        max_uptake (float): maximum uptake rate (default: 10)

    Returns:
        float: MRO score
    """

    exch_reactions = set(community.merged.get_exchange_reactions())
    max_uptake = max_uptake * len(community.organisms)

    if environment:
        environment.apply(community.merged, inplace=True, warning=False)
        exch_reactions &= set(environment)

    medium, sol = minimal_medium(community.merged, exchange_reactions=exch_reactions, direction=direction,
                                 min_mass_weight=min_mol_weight, min_growth=min_growth, max_uptake=max_uptake,
                                 validate=validate,  warnings=False, milp=(not use_lp))

    if sol.status != Status.OPTIMAL:
        if verbose:
            warn('MRO: Failed to find a valid solution for community')
        return None, None

    interacting_env = Environment.from_reactions(medium, max_uptake=max_uptake)
    interacting_env.apply(community.merged, inplace=True)

    if exclude is None:
        exclude = set()

    medium = {x[7:-7] for x in medium} - exclude
    individual_media = {}
    solver = solver_instance(community.merged)

    for org_id in community.organisms:
        biomass_reaction = community.organisms_biomass_reactions[org_id]
        community.merged.biomass_reaction = biomass_reaction
        org_interacting_exch = community.organisms_exchange_reactions[org_id]

        medium_i, sol = minimal_medium(community.merged, exchange_reactions=org_interacting_exch, direction=direction,
                                     min_mass_weight=min_mol_weight, min_growth=min_growth, max_uptake=max_uptake,
                                     validate=validate, solver=solver, warnings=False, milp=(not use_lp))

        if sol.status != Status.OPTIMAL:
            warn('MRO: Failed to find a valid solution for: ' + org_id)
            return None, None

        individual_media[org_id] = {org_interacting_exch[r].original_metabolite[2:-2] for r in medium_i} - exclude

    pairwise = {(o1, o2): individual_media[o1] & individual_media[o2] for o1, o2 in combinations(community.organisms, 2)}

    numerator = sum(map(len, pairwise.values())) / len(pairwise) if len(pairwise) != 0 else 0
    denominator = sum(map(len, individual_media.values())) / len(individual_media) if len(individual_media) != 0 else 0
    score = numerator / denominator if denominator != 0 else None

    extras = {
        'community_medium': medium,
        'individual_media': individual_media
    }

    return score, extras


def minimal_environment(community, aerobic=None, min_mol_weight=False, min_growth=0.1, max_uptake=10,
                        validate=False, verbose=True, use_lp=False):

    exch_reactions = set(community.merged.get_exchange_reactions())

    exch_reactions -= {"R_EX_M_h2o_e_pool"}
    community.merged.set_flux_bounds("R_EX_M_h2o_e_pool", -inf, inf)

    if aerobic is not None:
        exch_reactions -= {"R_EX_M_o2_e_pool"}
        if aerobic:
            community.merged.set_flux_bounds("R_EX_M_o2_e_pool", -max_uptake, inf)
        else:
            community.merged.set_flux_bounds("R_EX_M_o2_e_pool", 0, inf)

    ex_rxns, sol = minimal_medium(community.merged, exchange_reactions=exch_reactions,
        min_mass_weight=min_mol_weight, min_growth=min_growth, milp=(not use_lp),
        max_uptake=max_uptake, validate=validate, warnings=False)

    if ex_rxns is None:
        if verbose:
            warn('Failed to find a medium for interacting community.')
        return None
    else:
        if aerobic is not None and aerobic:
            ex_rxns |= {"R_EX_M_o2_e_pool"}
        env = Environment.from_reactions(ex_rxns, max_uptake=max_uptake)
        env["R_EX_M_h2o_e_pool"] = (-inf, inf)
        return env



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from smetana.interface import main
import pandas as pd


class TestGlobal(unittest.TestCase):
    def test_global(self):
        main(["tests/data/ec_*_ko.xml"], mode="global", output="tests/output/test", media="M9,LB",
             mediadb="tests/data/media_db.tsv", exclude="tests/data/inorganic.txt")
        df = pd.read_csv("tests/output/test_global.tsv")
        self.assertEqual(df.shape[0], 2)


class TestDetailed(unittest.TestCase):

    def test_detailed(self):
        main(["tests/data/ec_*.xml"], mode="detailed", output="tests/output/test", media="M9,LB",
             mediadb="tests/data/media_db.tsv", exclude="tests/data/inorganic.txt")
        df = pd.read_csv("tests/output/test_detailed.tsv")
        self.assertGreater(df.shape[0], 5)
        self.assertLess(df.shape[0], 15)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = ["reframed>=1.6.0", "pandas>=2.0.0"]

test_requirements = requirements + ['cplex']

script_list = [
    "scripts/smetana",
]


setup(
    author="Daniel Machado",
    author_email='cdanielmachado@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console', 
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: Apache Software License',
    ],
    description="Species METabolic interaction ANAlysis (SMETANA) is a python-based command line tool to analyse microbial communities.",
    scripts=script_list,
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    include_package_data=True,
    keywords='smetana',
    name='smetana',
    test_suite='tests',
    tests_require=test_requirements,
    packages=find_packages(include=['smetana']),
    setup_requires=['setuptools_scm'],
    url='https://github.com/cdanielmachado/smetana',
    version='1.2.1',
    zip_safe=False,
)



from collections import OrderedDict
from reframed.core.cbmodel import CBModel, CBReaction
from reframed.core.model import Compartment, Metabolite, ReactionType
from reframed.core.model import AttrOrderedDict
from warnings import warn
from copy import deepcopy
from math import inf


class CommunityNameMapping(object):
    def __init__(self, original_reaction=None, organism_reaction=None, original_metabolite=None,
                 organism_metabolite=None, extracellular_metabolite=None, community_exchange_reaction=None):
        """
        This class is used to represent mapping between original and merged community model metabolites and reactions

        Args:
            original_reaction (str): Name of reaction in original model
            organism_reaction (str): Name of reaction in merged community model
            original_metabolite (str): Name of metabolite in original model
            organism_metabolite (str): Name of metabolite in merged community model
            extracellular_metabolite (str): Name of "common environment" metabolite in merged community model
        """
        self.original_reaction = original_reaction
        self.organism_reaction = organism_reaction
        self.original_metabolite = original_metabolite
        self.organism_metabolite = organism_metabolite
        self.extracellular_metabolite = extracellular_metabolite
        self.community_exchange_reaction = community_exchange_reaction

    def __repr__(self):
        repr_str =  "<orig_m: {}, org_m: {}, ex_m: {}, orig_r: {}, org_r: {}, exch_r: {}>"
        return repr_str.format(self.original_metabolite,
                               self.organism_metabolite,
                               self.extracellular_metabolite,
                               self.original_reaction,
                               self.organism_reaction,
                               self.community_exchange_reaction)


class Community(object):
    """
    This class implements a microbial community model.

    It serves as a container for multiple organisms, and can be used to merge multiple single-species models (CBModel)
    into a single multi-species model (CBModel) which is compatible with most types of constraint-based methods.
    """

    def __init__(self, community_id, models=None, copy_models=True,
                 merge_extracellular_compartments=False, create_biomass=True, interacting=True,
                 exchanged_metabolites_blacklist=set()):
        """

        Args:
            community_id (str): community identifier
            models (list): list of models to be merged into single community
            copy_models (bool): If true copies for merged models  are created
            extracellular_compartment_id (str): Extracellular compartment id is used when merging extracellular compartments
            merge_extracellular_compartments (bool): Do not create organism specific extracellular compartment
            create_biomass (bool): create biomass reaction with biomass metabolites as reactants
            interacting (bool): If true models will be able to exchange metabolites. Otherwise all produced metabolites will go to sink
            exchanged_metabolites_blacklist (set): List of metabolites that can not be exchanged between species. This is done
             by separating 'pool' (uptake) compartment and 'pool' ('export') compartments for certain metabolites.
        """

        if not interacting and merge_extracellular_compartments:
            raise RuntimeError("Non-interacting models are not supported when merging extracellular compartment")

        self.id = community_id
        self._organisms = AttrOrderedDict()
        self._merge_extracellular_compartments = merge_extracellular_compartments
        self._create_biomass = create_biomass
        self._merged_model = None
        self._copy_models = copy_models
        self._interacting = interacting
        self._organisms_exchange_reactions = {}
        self._organisms_biomass_reactions = {}
        self._exchanged_metabolites_blacklist = set(exchanged_metabolites_blacklist)

        if models is not None:
            for model in models:
                self.add_organism(model, copy_models)

    @property
    def copy_models(self):
        """
        If true copies for merged models  are created

        Returns: bool
        """
        return self._copy_models

    @property
    def create_biomass_reaction(self):
        """
        Create biomass reaction with biomass metabolites as reactants

        Returns: bool
        """
        return self._create_biomass

    @create_biomass_reaction.setter
    def create_biomass_reaction(self, value):
        """
        Create biomass reaction with biomass metabolites as reactants

        Args:
            value: bool
        """
        self._clear_merged_model()
        self._create_biomass = value

    @property
    def size(self):
        return float(len(self._organisms))

    @property
    def interacting(self):
        """
        If true models will be able to exchange metabolites. Otherwise all produced metabolites will go to sink

        Returns: bool
        """
        return self._interacting

    @interacting.setter
    def interacting(self, value):
        """
        If true models will be able to exchange metabolites. Otherwise all produced metabolites will go to sink

        Args:
            value: bool
        """
        self._clear_merged_model()
        self._interacting = value

    @property
    def organisms_exchange_reactions(self):
        """
        Returns dictionary containing list of reactions exchanging model metabolites with common environment.
        Dictionary keys are model ids. Values are dictionaries with keys containing exchange reaction ids and values
        containing various information about these reactions.

        Returns: dict
        """
        if not self._merged_model:
            self._merged_model = self.generate_merged_model()

        return self._organisms_exchange_reactions

    @property
    def organisms_reactions(self):
        """
        Returns dictionary containing list of community organisms specific reactions

        Returns: dict
        """
        if not self._merged_model:
            self._merged_model = self.generate_merged_model()

        return self._organisms_reactions

    @property
    def organisms_biomass_reactions(self):
        """
        Returns dictionary containing reaction exporting biomass to common environment. Keys are model ids, and values
        are reaction ids

        Returns: dict
        """
        if not self._merged_model:
            self._merged_model = self.generate_merged_model()

        return self._organisms_biomass_reactions

    @property
    def merge_extracellular_compartments(self):
        """
        Do not create organism specific extracellular compartment

        Returns: bool
        """
        return self._merge_extracellular_compartments

    @merge_extracellular_compartments.setter
    def merge_extracellular_compartments(self, value):
        """
        Do not create organism specific extracellular compartment

        Args:
            value: bool
        """
        self._clear_merged_model()
        self._merge_extracellular_compartments = value

    @property
    def merged(self):
        """
        Merged models containing every organism as separate compartment

        Returns: CBModel
        """
        if not self._merged_model:
            self._merged_model = self.generate_merged_model()

        return self._merged_model

    @property
    def organisms(self):
        """
        Dictionary of organism models which are part of the community. Keys are model ids and values are models
        Returns: dict
        """
        return self._organisms

    def __str__(self):
        return '\n'.join(self._organisms.keys())

    def _clear_merged_model(self):
        self._merged_model = None
        self._organisms_exchange_reactions = {}
        self._organisms_reactions = {}

    def add_organism(self, model, copy=True):
        """ Add an organism to this community.

        Args:
            model (CBModel): model of the organism
            copy (bool): create a copy of the given model (default: True)

        """
        self._clear_merged_model()

        if model.id in self._organisms:
            warn("Organism '{}' is already in this community".format(model.id))
        else:
            if copy:
                model = model.copy()

            self._organisms[model.id] = model

    def remove_organism(self, organism):
        """ Remove an organism from this community

        Args:
            organism (str): organism id

        """
        self._clear_merged_model()

        if organism not in self._organisms:
            warn('Organism {} is not in this community'.format(organism))
        else:
            del self._organisms[organism]

    def generate_merged_model(self):
        def _id_pattern(object_id, organism_id):
            return "{}_{}".format(object_id, organism_id)

        def _name_pattern(object_name, organism_name):
            return "{} ({})".format(object_name, organism_name)

        def _copy_object(obj, org_id, compartment=None):
            new_obj = deepcopy(obj)
            new_obj.id = _id_pattern(obj.id, org_id)
            new_obj.name = _name_pattern(obj.name, org_id)
            if compartment:
                new_obj.compartment = compartment

            return new_obj

        models_missing_biomass = [m.id for m in self._organisms.values() if not m.biomass_reaction]
        if models_missing_biomass:
            raise RuntimeError("Biomass reaction not found in models: {}".format("', '".join(models_missing_biomass)))

        merged_model = CBModel(self.id)

        organisms_biomass_metabolites = {}
        community_metabolite_exchange_lookup = {}

        for org_id, model in self._organisms.items():
            self._organisms_reactions[org_id] = []
            self._organisms_exchange_reactions[org_id] = {}
            self._organisms_biomass_reactions[org_id] = {}
            exchanged_metabolites = {m_id for r_id in model.get_exchange_reactions()
                                     for m_id in model.reactions[r_id].stoichiometry}
            #
            # Create additional extracellular compartment
            #
            if not self._merge_extracellular_compartments:
                pool_compartment = Compartment('pool', 'common pool')
                merged_model.add_compartment(pool_compartment)
                export_pool_compartment = Compartment('pool_blacklist', 'blacklisted metabolite pool')
                merged_model.add_compartment(export_pool_compartment)

            for c_id, comp in model.compartments.items():
                if not comp.external or not self._merge_extracellular_compartments:
                    new_comp = _copy_object(comp, org_id)
                    merged_model.add_compartment(new_comp)
                elif c_id not in merged_model.compartments:
                    merged_model.add_compartment(deepcopy(comp))

            for m_id, met in model.metabolites.items():
                if not model.compartments[met.compartment].external or not self._merge_extracellular_compartments:
                    new_met = _copy_object(met, org_id, _id_pattern(met.compartment, org_id))
                    merged_model.add_metabolite(new_met)
                elif m_id not in merged_model.metabolites:
                    merged_model.add_metabolite(deepcopy(met))

                m_blacklisted = met.id in self._exchanged_metabolites_blacklist

                if met.id in exchanged_metabolites and not self._merge_extracellular_compartments:
                    #
                    # For blacklisted metabolites create a separate pool from which metabolites can not be reuptaken
                    #
                    if m_blacklisted and self._interacting:
                        pool_id = _id_pattern(m_id, "pool_blacklist")
                        if pool_id not in merged_model.metabolites:
                            new_met = _copy_object(met, "pool_blacklist", "pool_blacklist")
                            merged_model.add_metabolite(new_met)

                            exch_id = _id_pattern("R_EX_" + m_id, "pool_blacklist")
                            exch_name = _name_pattern(met.name, "pool (blacklist) exchange")
                            blk_rxn = CBReaction(exch_id, name=exch_name, reversible=False,
                                                 reaction_type=ReactionType.SINK)
                            blk_rxn.stoichiometry[pool_id] = -1.0
                            community_metabolite_exchange_lookup[new_met.id] = exch_id
                            merged_model.add_reaction(blk_rxn)

                    pool_id = _id_pattern(m_id, "pool")
                    if pool_id not in merged_model.metabolites:
                        new_met = _copy_object(met, "pool", "pool")
                        merged_model.add_metabolite(new_met)

                        exch_id = _id_pattern("R_EX_" + m_id, "pool")
                        exch_name = _name_pattern(met.name, "pool exchange")
                        new_rxn = CBReaction(exch_id, name=exch_name, reversible=True,
                                             reaction_type=ReactionType.EXCHANGE)
                        new_rxn.stoichiometry[pool_id] = -1.0
                        community_metabolite_exchange_lookup[new_met.id] = exch_id
                        merged_model.add_reaction(new_rxn)

            for r_id, rxn in model.reactions.items():

                is_exchange = rxn.reaction_type == ReactionType.EXCHANGE

                if not is_exchange or not self._merge_extracellular_compartments:
                    new_rxn = _copy_object(rxn, org_id)

                    for m_id, coeff in rxn.stoichiometry.items():
                        m_blacklisted = m_id in self._exchanged_metabolites_blacklist
                        if (not model.compartments[model.metabolites[m_id].compartment].external
                                or not self._merge_extracellular_compartments):
                            del new_rxn.stoichiometry[m_id]
                            new_id = _id_pattern(m_id, org_id)
                            new_rxn.stoichiometry[new_id] = coeff

                        if is_exchange:
                            new_rxn.reaction_type = ReactionType.OTHER
                            if (model.compartments[model.metabolites[m_id].compartment].external
                                    and not self._merge_extracellular_compartments):
                                # TODO: if m_id in self._exchanged_metabolites_blacklist:
                                pool_id = _id_pattern(m_id, "pool")
                                new_rxn.stoichiometry[pool_id] = -coeff
                                cnm = CommunityNameMapping(
                                    organism_reaction=new_rxn.id,
                                    original_reaction=r_id,
                                    organism_metabolite=new_id,
                                    extracellular_metabolite=pool_id,
                                    original_metabolite=m_id,
                                    community_exchange_reaction=community_metabolite_exchange_lookup[pool_id])
                                self._organisms_exchange_reactions[org_id][new_rxn.id] = cnm

                                if not self.interacting:
                                    sink_rxn = CBReaction('Sink_{}'.format(new_id), reaction_type=ReactionType.SINK,
                                                          reversible=False)
                                    sink_rxn.stoichiometry = {new_id: -1}
                                    sink_rxn.lb = 0.0
                                    merged_model.add_reaction(sink_rxn)
                                elif m_blacklisted:
                                    pool_blacklist_id = _id_pattern(m_id, "pool_blacklist")
                                    blacklist_export_rxn = CBReaction('R_EX_BLACKLIST_{}'.format(new_id),
                                                                      reaction_type=ReactionType.OTHER,
                                                                      reversible=False)
                                    blacklist_export_rxn.stoichiometry = {new_id: -1, pool_blacklist_id: 1}
                                    blacklist_export_rxn.lb = 0.0
                                    merged_model.add_reaction(blacklist_export_rxn)

                    if is_exchange and not self._merge_extracellular_compartments:
                        new_rxn.reversible = True
                        new_rxn.lb = -inf
                        new_rxn.ub = inf if self.interacting and not m_blacklisted else 0.0

                    if rxn.id == model.biomass_reaction:
                        new_rxn.reversible = False

                    if self._create_biomass and rxn.id == model.biomass_reaction:
                        new_rxn.objective = False

                        # Add biomass metabolite to biomass equation
                        m_id = _id_pattern('Biomass', org_id)
                        name = _name_pattern('Community biomass', org_id)
                        comp = 'pool'
                        biomass_met = Metabolite(m_id, name, comp)
                        merged_model.add_metabolite(biomass_met)
                        new_rxn.stoichiometry[m_id] = 1
                        organisms_biomass_metabolites[org_id] = m_id

                        sink_rxn = CBReaction('Sink_biomass_{}'.format(org_id), reaction_type=ReactionType.SINK,
                                              reversible=False)
                        sink_rxn.stoichiometry = {m_id: -1}
                        sink_rxn.lb = 0.0
                        merged_model.add_reaction(sink_rxn)

                    self._organisms_reactions[org_id].append(new_rxn.id)
                    merged_model.add_reaction(new_rxn)

                else:
                    if is_exchange and self._merge_extracellular_compartments:
                        self._organisms_exchange_reactions[org_id][rxn.id] = CommunityNameMapping(
                            organism_reaction=r_id,
                            original_reaction=r_id,
                            extracellular_metabolite=list(rxn.stoichiometry.keys())[0],
                            original_metabolite=list(rxn.stoichiometry.keys())[0],
                            organism_metabolite=None)
                        self._organisms_reactions[org_id].append(rxn.id)

                    if r_id in merged_model.reactions:
                        continue

                    new_rxn = deepcopy(rxn)
                    new_rxn.reaction_type = ReactionType.EXCHANGE
                    if rxn.id == model.biomass_reaction and self._create_biomass:
                        new_rxn.reversible = False
                        new_rxn.objective = False

                        m_id = _id_pattern('Biomass', org_id)
                        name = _name_pattern('Biomass', org_id)
                        comp = 'pool'
                        biomass_met = Metabolite(m_id, name, comp)
                        merged_model.add_metabolite(biomass_met)
                        new_rxn.stoichiometry[m_id] = 1
                        organisms_biomass_metabolites[org_id] = m_id

                    merged_model.add_reaction(new_rxn)

                if r_id == model.biomass_reaction:
                    self._organisms_biomass_reactions[org_id] = new_rxn.id

        if self._create_biomass:
            biomass_rxn = CBReaction('R_Community_Growth', name="Community Growth",
                                     reversible=False, reaction_type=ReactionType.SINK, objective=1.0)
            for org_biomass in organisms_biomass_metabolites.values():
                biomass_rxn.stoichiometry[org_biomass] = -1

            merged_model.add_reaction(biomass_rxn)
            merged_model.biomass_reaction = biomass_rxn.id

        return merged_model

    def copy(self, merge_extracellular_compartments=None, copy_models=None, interacting=None, create_biomass=None,
             exchanged_metabolites_blacklist=None):
        """
        Copy model object
        Args:
            copy_models (bool): If true copies for merged models  are created
            interacting (bool): If true models will be able to exchange metabolites. Otherwise all produced metabolites will go to sink
            create_biomass (bool): create biomass reaction with biomass metabolites as reactants
        Returns:
            Community
        """
        if copy_models is None:
            copy_models = self._copy_models

        if interacting is None:
            interacting = self._interacting

        if create_biomass is None:
            create_biomass = self._create_biomass

        if merge_extracellular_compartments is None:
            merge_extracellular_compartments = self._merge_extracellular_compartments

        if exchanged_metabolites_blacklist is None:
            exchanged_metabolites_blacklist = self._exchanged_metabolites_blacklist

        copy_community = Community(self.id, models=list(self._organisms.values()),
                                   copy_models=copy_models, create_biomass=create_biomass,
                                   merge_extracellular_compartments=merge_extracellular_compartments,
                                   interacting=interacting,
                                   exchanged_metabolites_blacklist=exchanged_metabolites_blacklist)

        return copy_community

    def split_fluxes(self, fluxes):
        """ Decompose a flux balance solution of the merged community into organism-specific flux vectors.

        Args:
            fluxes (dict): flux distribution as a single dict

        Returns:
            dict: community flux distribution as a nested dict
        """

        comm_fluxes = OrderedDict()

        for org_id, model in self._organisms.items():
            org_fluxes = [(r_id[:-(1 + len(org_id))], val) for r_id, val in fluxes.items() if r_id.endswith(org_id)]
            comm_fluxes[org_id] = OrderedDict(org_fluxes)

        return comm_fluxes